'''Module webapp api wrapper'''
import requests
# App imports
from .abstract import Abstract
from basepair.helpers import eprint

class Module(Abstract):
  '''Webapp Module class'''
  def __init__(self, cfg):
    super(Module, self).__init__(cfg)
    self.api_endpoint = '{}pipelines/get_module'.format(self.endpoint)
    self.endpoint += 'modules/'

  def get_pipeline_modules(self, obj_id, cache=False, params={}, verify=True): # pylint: disable=dangerous-default-value
    '''Get modules of an pipeline'''
    params.update(self.payload)
    try:
      response = requests.get(
        '{}?workflow={}'.format(self.api_endpoint, obj_id),
        params=params,
        verify=verify,
      )
      parsed = self._parse_obj_response(response, obj_id)
      return parsed.get('objects')
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}



"""Driver for Opsgenie"""

# Libs import
from opsgenie_sdk import AlertApi, CreateAlertPayload
from opsgenie_sdk.api_client import ApiClient
from opsgenie_sdk.configuration import Configuration

# App imports
from .abstract import AlertAbstract


class Instance(AlertAbstract):
	"""OPSGenie Implementation of Abstract class"""

	alert_api = None
	api_client = None

	def critical(self, msg, payload):
		"""Send critical severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, "P1"))

	def high(self, msg, payload):
		"""Send high severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, "P2"))

	def info(self, msg, payload):
		"""Send very low severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, "P5"))

	def low(self, msg, payload):
		"""Send low severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, "P4"))

	def medium(self, msg, payload):
		"""Send medium severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, "P3"))

	def set_config(self, cfg):
		self.cfg = Configuration()
		self.cfg.api_key["Authorization"] = cfg.get("secret_key")
		self.api_client = ApiClient(configuration=self.cfg)
		self.alert_api = AlertApi(api_client=self.api_client)

	def _parse_payload(self, msg, payload, priority="P3"):
		"""Helper to parse payload"""
		payload = payload or {}
		payload["priority"] = priority
		payload["message"] = msg
		return payload

	def _send_alert(self, payload):
		"""Helper to send alerts"""
		return self.alert_api.create_alert(
			create_alert_payload=CreateAlertPayload(**payload)
		)



'''AWS module'''
from .cw import CW
from .ec2 import EC2
from .efs import EFS
from .handler.exception import ExceptionHandler
from .hos import HOS
from .how import HOW
from .iam import IAM
from .instance import INSTANCE_INFO
from .logs import Logs
from .mrktpl import MMetering
from .policy import Policy
from .s3 import S3
from .service import Service
from .sm import SM
from .sqs import SQS
from .sts import STS
from .swf import SWF



'''Upload webapp api wrapper'''

# General imports
import json

# Lib imports
import requests

# App imports
from basepair.helpers import eprint
from .abstract import Abstract

class Upload(Abstract):
  '''Webapp Upload class'''
  def __init__(self, cfg):
    super(Upload, self).__init__(cfg)
    self.endpoint += 'uploads/'

  def bulk_import(self, payload={}, verify=True):
    '''Import sample from s3'''
    try:
      response = requests.post(
        '{}bulk_import'.format(self.endpoint),
        data=json.dumps(payload),
        headers=self.headers,
        params=self.payload,
        verify=verify,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}



'''User webapp api wrapper'''

# General imports
import json
import os

# Lib imports
import requests

# App imports
from basepair.helpers import eprint
from .abstract import Abstract

class User(Abstract):
  '''Webapp User class'''
  def __init__(self, cfg):
    super(User, self).__init__(cfg)
    self.endpoint += 'users/'

  def get_configuration(self, cache=False, verify=True):
    '''Get host configuration for user'''
    if cache:
      filename = os.path.expanduser(cache)
      if os.path.exists(filename) and os.path.getsize(filename):
        return json.loads(open(filename, 'r').read().strip())

    params = {'origin': 'cli'}
    params.update(self.payload)
    try:
      response = requests.get(
        '{}get_configuration'.format(self.endpoint),
        params=params,
        verify=verify,
      )
      parsed = self._parse_response(response)

      # save in cache if required
      if cache and parsed and not parsed.get('error'):
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
          os.makedirs(directory)
        eprint('Saving configuration into:', filename)
        with open(filename, 'w') as handle:
          handle.write(json.dumps(parsed, indent=2))
      return parsed
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}



'''Common parser used in datatypes parsing'''
import argparse
import os
import re
import sys

# App imports
from basepair.helpers import eprint

def add_common_args(parser):
  '''Add common args'''
  parser.add_argument('-c', '--config', help='Path to API config file')
  parser.add_argument('--quiet', action='store_true')
  parser.add_argument('--cache-cloud-credential', dest='keep_cloud_service_conf', action='store_true')
  parser.add_argument('--use-cache', action='store_true')
  parser.add_argument('--scratch', default='.', help='Scratch dir for files')
  parser.add_argument('--verbose', action='store_true')
  return parser

def add_payload_args(parser):
  '''Add payload args'''
  parser.add_argument(
    '--payload-username',
    help='Replace payload api key. Must also specify --payload-api-key'
  )
  parser.add_argument(
    '--payload-api-key',
    help='Replace payload api key. Must also specify --payload-username'
  )
  return parser

def add_force_parser(parser, datatype):
  '''Add force parser'''
  parser.add_argument(
    '--force',
    action='store_true',
    help='(Optional) Override existing {}.'.format(datatype)
  )
  return parser

def add_json_parser(parser):
  '''Add json parser'''
  parser.add_argument(
    '--json',
    action='store_true',
    help='(Optional) Print the data in JSON format (this shows all data associated with each item).'
  )
  return parser

def add_outdir_parser(parser):
  '''Add outdir parser'''
  parser.add_argument(
    '-o',
    '--outdir',
    required=False,
    default='.',
    help='Base directory to save files to (default \'.\').'
  )
  return parser

def add_single_uid_parser(parser,datatype):
  '''Add single uid parser'''
  parser.add_argument(
    '-u',
    '--uid',
    help='The unique id for {}'.format(datatype),
    required=True,
    type=valid_uid
  )
  return parser

def add_pid_parser(parser):
  '''Add pipeline id parser'''
  parser.add_argument(
    '--pipeline',
    help='A unique pipeline id',
    required=True,
    type=valid_uid
  )
  return parser

def add_tags_parser(parser):
  '''Add tags parser'''
  parser.add_argument(
    '--tags',
    nargs='+',
    action='append',
    default=None
  )
  parser.add_argument(
    '--tagkind',
    choices=['diff', 'exact', 'subset'],
    default='exact',
    help='''
             Tag filtering strategy. Options:\n
             (1) exact (default): only get files with exactly these tags.
             (2) diff: exclude files with this tag.
             (3) subset: any files with one or more of these tags.
             '''
  )
  return parser

def add_uid_parser(parser, datatype):
  '''Add uid parser'''
  parser.add_argument(
    '-u',
    '--uid',
    default=None,
    help='The unique id(s) for {}'.format(datatype),
    nargs='+',
    required=True,
    type=valid_uid
  )
  return parser

def add_yaml_parser(parser):
  '''Add yaml file parser'''
  parser.add_argument(
    '--file',
    default=None,
    required=False,
    help='The filepath of YAML',
    nargs='+'
  )
  return parser

def valid_uid(value):
  '''Validates the uid for positive integer'''
  if value.isdigit() and int(value) > 0:
    return value
  raise argparse.ArgumentTypeError('ERROR: uid must be a positive integer')

def valid_email(value):
  '''Validates the email'''
  pattern = re.compile(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$")
  if pattern.match(value):
    return value
  raise argparse.ArgumentTypeError('ERROR: Invalid email format.')

valid_sample_extensions = ('.ab1', '.bai', '.bam', '.crai', '.csfasta', '.csv', '.fasta', '.fastq', '.fq', '.gvcf', '.qual', '.vcf', '.sam', '.sra', '.txt', '.bz', '.bz2', '.gz', '.zip')

def validate_sample_file(files):
  '''Validates sample file type'''
  for file in files:
    if not file.endswith(valid_sample_extensions):
      sys.exit('ERROR: Please provide valid sample file. Available File types - {}'.format(' '.join(valid_sample_extensions)))
  return True

def validate_conf(args):
  '''Helper to validate the proper configuration argument is being set'''
  if args.config:
    return eprint('Using config file', args.config)
  if 'BP_CONFIG_FILE' in os.environ:
    return eprint('Using config file {}'.format(os.environ['BP_CONFIG_FILE']))
  return sys.exit('ERROR: Please either use the -c or --config param or set the environment variable BP_CONFIG_FILE!')

def validate_analysis_yaml(yaml_argument):
  '''Checks yaml file'''
  if not isinstance(yaml_argument, list):
    yaml_argument = [yaml_argument]
  for each_yaml in yaml_argument:
    valid_extensions = ('.yaml', '.yml')
    if not each_yaml.endswith(valid_extensions):
      sys.exit('ERROR: Please provide yaml file with extension .yaml, .yml ')
    elif not os.path.isfile(each_yaml):
      sys.exit('ERROR: File does not exist at {}.'.format(each_yaml))
  return True

def validate_update_yaml(args):
  '''Validates yaml file'''
  if not args.file:
    sys.exit('ERROR: Yaml file required.')
  if len(args.file) != 1:
    sys.exit('ERROR: Please provide only one yaml')
  yaml_path = args.file[0]
  valid_extensions = ('.yaml', '.yml')
  if not yaml_path.endswith(valid_extensions):
    sys.exit('ERROR: Please provide yaml file with extension .yaml, .yml ')
  elif not os.path.isfile(yaml_path):
    sys.exit('ERROR: File does not exist at {}.'.format(yaml_path))
  return True

def validate_create_yaml(args):
  '''Validates yaml file'''
  if not args.file:
    sys.exit('ERROR: Yaml file required.')
  for i in range(0, len(args.file)):
    yaml_path = args.file[i]
    valid_extensions = ('.yaml', '.yml')
    if not yaml_path.endswith(valid_extensions):
      sys.exit('ERROR: Please provide yaml file with extension .yaml, .yml ')
    elif not os.path.isfile(yaml_path):
      sys.exit('ERROR: File does not exist at {}.'.format(yaml_path))
  return True



'''Helper to filter sets of data'''

class SetFilter:
  '''Helper class to filter list'''
  @staticmethod
  def diff(a_set, b_set):
    '''Filter by not intersection'''
    return not set(b_set).intersection(set(a_set))

  @staticmethod
  def exact(a_set, b_set):
    '''Filter by eq'''
    return set(a_set) == set(b_set)

  @staticmethod
  def subset(a_set, b_set):
    '''Filter by intersection'''
    return set(a_set).intersection(set(b_set))



'''File datatype class'''

# General Import
import sys

# App imports
from bin.common_parser import add_uid_parser, add_common_args, add_outdir_parser

class File:
  '''File action methods'''

  @staticmethod
  def download_file(bp_api, args):
    '''Download file by uid'''
    all_fail = True
    for uid in args.uid:
      file_i = bp_api.get_file(uid)
      all_fail = not (bool(file_i.get('id')) \
        and bool(bp_api.download_file(file_i['path'], uid=uid, dirname=args.outdir, file_type='files'))) \
        and all_fail
    if all_fail:
      sys.exit('ERROR: File downloading failed.')

  @staticmethod
  def file_action_parser(action_parser):
    '''File datatype action parser'''
    download_file_p = action_parser.add_parser(
      'download',
      help='Download one or more files by uid.'
    )
    download_file_p = add_uid_parser(download_file_p, 'file')
    download_file_p = add_outdir_parser(download_file_p)
    download_file_p = add_common_args(download_file_p)

    return action_parser



'''AWS market place metering'''
# App imports
from botocore.exceptions import ClientError
from basepair.modules.aws.service import Service

class MMetering(Service):
  '''Wrapper for Marketplace services'''
  def __init__(self, cfg):
    super().__init__(cfg, service_name='Marketplace metering')
    self.client = self.session.client(**{'service_name': 'meteringmarketplace'})
    self.product_code = cfg.get('product_code')

  def resolve_customer(self, registration_token):
    '''resolve customer method'''
    try:
      response = self.client.resolve_customer(RegistrationToken=registration_token)
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to resolve customer with token - {registration_token}.',
      })
      raise error
    return response

  def send_batch_credit_usage(self, usage_record):
    '''send credit usage records method'''
    try:
      response = self.client.batch_meter_usage(UsageRecords=usage_record, ProductCode=self.product_code)
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to send credit usage records - {str(usage_record)}',
      })
      raise error
    return response



'''Pipeline webapp api wrapper'''

# App imports
from .abstract import Abstract

class Pipeline(Abstract):
  '''Webapp Pipeline class'''
  def __init__(self, cfg):
    super(Pipeline, self).__init__(cfg)
    self.endpoint += 'pipelines/'



from datetime import datetime

from basepair.modules.identity.drivers import IdentityAbstract


class Driver(IdentityAbstract):
  def __init__(self, cfg=None):
    self.iam_settings = cfg.get('settings', {})
    self.iam_credentials = cfg.get('credentials', {})
    self.username_pattern = cfg.get('username_pattern') or '_tmp_{}'

  def delete_access_keys(self, username):
    '''Method to delete access keys'''
    pass

  def delete_inline_policies(self, username):
    '''Method to delete delete inline policies'''
    pass

  def delete_policy(self, policy_arn):
    '''Method to delete policy'''
    pass

  def get_policy_versions(self, policy_arn):
    '''Method to get policy version'''
    pass

  def delete_user(self, uid, ignore_format=False):
    '''Method to delete user'''
    pass

  def detach_policies(self, username, delete=True):
    '''Method to detach policies'''
    pass

  def setup_user(self, uid, boundary='', tags=[]):
    '''Method to create a user'''
    # return username, access_key
    username = self.username_pattern.format(uid)
    credentials = {
      'AccessKeyId': self.iam_credentials.get('id'),
      'CreateDate': datetime.now().strftime('%m/%d/%Y, %H:%M:%S'),
      'SecretAccessKey': self.iam_credentials.get('secret'),
      'Status': 'Active',
      'UserName': username,
    }
    return username, credentials

  def set_inline_policy(self, policy_scheme, username, replace=True):
    '''Method to set inline policy'''
    pass

  def set_policy(self, policy_scheme, username, force=True, sufix='', tags=[]):
    '''Method to set policy'''
    pass



"""AWS HealthOmics Workflow Wrapper"""

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.service import Service


class HOW(Service):
	"""Wrapper for Health Omics Workflow services"""

	def __init__(self, cfg):
		super().__init__(cfg, "HOW")
		self.client = self.session.client(
			**{
				"config": Config(retries={"max_attempts": 0, "mode": "standard"}),
				"service_name": "omics",
			}
		)

	def create_workflow(self, params):
		"""Create workflow"""
		try:
			response = self.client.create_workflow(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to create HealthOmics workflow: {str(error)}.",
				}
			)
			raise error
		return response

	def create_workflow_version(self, params):
		"""Create workflow version"""
		try:
			response = self.client.create_workflow_version(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to create HealthOmics workflow version: {str(error)}.",
				}
			)
			raise error
		return response

	def update_workflow(self, params):
		"""Update workflow"""
		try:
			self.client.update_workflow(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to update HealthOmics workflow: {str(error)}.",
				}
			)
			raise error
		return {"status": "success", "message": "Workflow updated successfully"}

	def update_workflow_version(self, params):
		"""Update workflow version"""
		try:
			self.client.update_workflow_version(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to update HealthOmics workflow version: {str(error)}.",
				}
			)
			raise error
		return {"status": "success", "message": "Workflow version updated successfully"}

	def cancel_run(self, run_id):
		"""Cancel run"""
		try:
			response = self.client.cancel_run(id=run_id)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to cancel HealthOmics run: {str(error)}.",
				}
			)
			raise error
		return response

	def delete_run(self, run_id):
		"""Delete run"""
		try:
			response = self.client.delete_run(id=run_id)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to delete HealthOmics run: {str(error)}.",
				}
			)
			raise error
		return response

	def delete_workflow(self, params):
		"""Delete workflow"""
		try:
			response = self.client.delete_workflow(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to delete HealthOmics workflow: {str(error)}.",
				}
			)
			raise error
		return response

	def get_run(self, params):
		"""Get omics run"""
		try:
			return self.client.get_run(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f'Not able to Get HealthOmics run id {params.get("id")}: {str(error)}.',
				}
			)
			raise error

	def get_run_task(self, params):
		"""Get omics run task"""
		try:
			return self.client.get_run_task(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f'Not able to Get HealthOmics run task id {params.get("id")}: {str(error)}.',
				}
			)
			raise error

	def get_workflow(self, params):
		"""Get omics workflow"""
		try:
			return self.client.get_workflow(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f"Not able to get HealthOmics workflow: {str(error)}.",
				}
			)
			raise error

	def get_workflow_version(self, params):
		"""Get omics workflow version"""
		try:
			return self.client.get_workflow_version(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f"Not able to get HealthOmics workflow version: {str(error)}.",
				}
			)
			raise error

	def list_run_tasks(self, params):
		"""Get omics run list"""
		try:
			tasks = []
			while True:
				response = self.client.list_run_tasks(**params)
				next_token = response.get("nextToken")
				tasks += response.get("items")
				if next_token:
					params["startingToken"] = next_token
				else:
					break
			return tasks
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f'Not able to Get HealthOmics run task list id {params.get("id")}: {str(error)}',
				}
			)
			raise error

	def list_workflows(self, params):
		"""List omics workflows"""
		try:
			return self.client.list_workflows(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f"Not able to list HealthOmics workflows: {str(error)}.",
				}
			)
			raise error

	def start_run(self, params):
		"""Start omics workflow"""
		try:
			return self.client.start_run(**params)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": str(error),
					"msg": f'Not able to start HealthOmics workflow id {params.get("workflowId")}: {str(error)}.',
				}
			)
			raise error



'''AWS EFS Wrappers'''

# General import
import hashlib
import time
from datetime import datetime

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class EFS(Service): # pylint: disable=too-few-public-methods
	'''Wrapper for EFS services'''
	def __init__(self, cfg):
		super().__init__(cfg, 'EFS')
		self.client = self.session.client(**{
			'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
			'service_name': 'efs',
		})

	def create_file_system(self, options):
		'''Create a file system'''
		options['CreationToken'] = hashlib.md5(f'{datetime.now()}'.encode()).hexdigest()
		options['PerformanceMode'] = options.get('PerformanceMode') or 'generalPurpose'
		options['Encrypted'] = True
		options['ThroughputMode'] = options.get('ThroughputMode') or 'bursting'
		states = ['creating', 'updating']
		try:
			response = self.client.create_file_system(**options)
			check = response.get('LifeCycleState') in states
			while check:
				time.sleep(10)
				check_response = self.client.describe_file_systems(
					CreationToken=options['CreationToken'],
					MaxItems=1,
				)
				file_system = next(iter(check_response.get('FileSystems')), None)
				check = file_system.get('LifeCycleState') in states
		except ClientError as error:
			response = self.get_log_msg({
				'exception': error,
				'msg': 'Not able to create file system.',
			})
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return response

	def create_mount_target(self, file_system_id):
		'''create a moun target'''
		options = {
			'FileSystemId': file_system_id,
			'SecurityGroups': self.cfg.get('security_groups'),
		}
		for subnet_id in self.cfg.get('subnet_ids'):
			options['SubnetId'] = subnet_id
			try:
				response = self.client.create_mount_target(**options)
			except ClientError as error:
				response = self.get_log_msg({
					'exception': error,
					'msg': f'Not able to create mount target for given file system id {file_system_id}.',
				})
				if ExceptionHandler.is_throttled_error(exception=error):
					raise error
			if response.get('IpAddress'):
				break
		return response

	def delete(self, file_system_id):
		'''Delete a file system and all its mount targets'''
		# get all mount targets to delete them
		mount_targets = self.__get_all_mount_targets(file_system_id)
		try:
			for mount_target in mount_targets:
				self.client.delete_mount_target(MountTargetId=mount_target.get('MountTargetId'))
		except ClientError as error:
			self.get_log_msg({
				'exception': error,
				'msg': f'Not able to remove the mount targets for the given file system id {file_system_id}.',
			})
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error

		while True:
			mount_targets = self.__get_all_mount_targets(file_system_id)
			if mount_targets:
				time.sleep(10)
				continue
			break

		# delete the file system
		try:
			self.client.delete_file_system(FileSystemId=file_system_id)
		except ClientError as error:
			self.get_log_msg({
				'exception': error,
				'msg': f'Not able to remove the given file system id {file_system_id}.',
			})
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error

	def get_by_tags(self, tags):
		'''Helper to get the file systems by tag'''
		try:
			file_systems = self.__get_all_file_systems()
			for file_system in file_systems:
				could_be_this_one = True
				for file_system_tag in file_system.get('Tags'):
					key, value = file_system_tag.get('key'), file_system_tag.get('value')
					could_be_this_one = could_be_this_one and tags.get(key) == value
					if not could_be_this_one:
						break
				if could_be_this_one:
					return file_system
		except Exception as error: # pylint: disable=broad-except
			return self.get_log_msg({
				'exception': error,
				'msg': 'Not able to retrieve file systems to filter by tags',
			})
		return None

	# Private method
	def __get_full_list(self, arguments, error_msg, field, method):
		'''Helper to get full list of item of a given API method'''
		response = []
		while True:
			try:
				next_response = getattr(self.client, method)(**arguments)
				next_marker = next_response.get('NextMarker')
				arguments['Marker'] = next_marker
				response += next_response.get(field)
				should_exit = not bool(next_marker) or bool(next_response.get(field)) or len(next_response.get(field)) < 100
				if should_exit:
					break
			except ClientError as error:
				response = self.get_log_msg({
					'exception': error,
					'msg': error_msg,
				})
				if ExceptionHandler.is_throttled_error(exception=error):
					raise error
		return response


	def __get_all_file_systems(self):
		'''Retrieve the full list of file system'''
		return self.__get_full_list(
			arguments={},
			error_msg='Not able to retrieve all the file system.',
			field='FileSystems',
			method='describe_file_systems',
		)

	def __get_all_mount_targets(self, file_system_id):
		'''Retrieve the full list of mount targets'''
		return self.__get_full_list(
			arguments={'FileSystemId': file_system_id},
			error_msg=f'Not able to retrieve all the mount targets for the given file system id {file_system_id}.',
			field='MountTargets',
			method='describe_mount_targets',
		)



def raise_no_implemented():
  '''Helper to raise exception when call no implemented method'''
  raise Exception('Abstract method require to be implemented.')

class SecretsAbstract:
  '''Abstract Driver class'''

  def __init__(self, cfg=None):
    '''Abstract secret constructor'''
    raise Exception('Abstract class cannot be instantiated.')

  def get(self, secret_id, use_cache):
    '''Method to get the secrets'''
    raise_no_implemented()



'''Module datatype class'''

# General Import
import sys

# App imports
from bin.common_parser import add_common_args, add_uid_parser, add_json_parser, add_yaml_parser, add_pid_parser, add_force_parser, validate_create_yaml, validate_update_yaml

class Module:
  '''Module action methods'''

  @staticmethod
  def create_module(bp_api, args):
    '''Create module'''
    valid = validate_create_yaml(args)
    if valid:
      for each_yaml in args.file:
        bp_api.create_module({'yamlpath': each_yaml, 'force': args.force})

  @staticmethod
  def delete_module(bp_api, args):
    '''Delete Module'''
    all_fail = True
    for uid in args.uid:
      answer = bp_api.yes_or_no('Are you sure you want to delete {}?'.format(uid))
      if answer:
        all_fail = not bool(bp_api.delete_module(uid)) and all_fail
    if all_fail:
      sys.exit('ERROR: Deleting sample failed.')

  @staticmethod
  def get_module(bp_api, args):
    '''Get module'''
    all_fail = True
    for uid in args.uid:
      all_fail = not bool(bp_api.print_data(data_type='module', uid=uid, is_json=args.json)) and all_fail
    if all_fail:
      sys.exit('ERROR: Module data not found.')

  @staticmethod
  def list_module(bp_api, args):
    '''List Modules'''
    result = bp_api.print_data(data_type='pipeline_modules', uid=args.pipeline, is_json=args.json)
    if not result:
      sys.exit('ERROR: Module data not found.')

  @staticmethod
  def update_module(bp_api, args):
    '''Update module'''
    valid = validate_update_yaml(args)
    if valid:
      bp_api.update_module({'yamlpath': args.file[0]})

  @staticmethod
  def module_action_parser(action_parser):
    '''Module parser'''

    # create module parser
    create_module_p = action_parser.add_parser(
      'create',
      help='Create module from yaml.'
    )
    create_module_p = add_common_args(create_module_p)
    create_module_p = add_yaml_parser(create_module_p)
    create_module_p = add_force_parser(create_module_p, 'module')

    # delete module parser
    delete_module_p = action_parser.add_parser(
      'delete',
      help='delete a module.'
    )
    delete_module_p = add_common_args(delete_module_p)
    delete_module_p = add_uid_parser(delete_module_p, 'module')

    # get module parser'''
    get_module_p = action_parser.add_parser(
      'get',
      help='Get details of a module'
    )
    get_module_p = add_common_args(get_module_p)
    get_module_p = add_uid_parser(get_module_p, 'module')
    get_module_p = add_json_parser(get_module_p)

    # list module parser
    list_modules_p = action_parser.add_parser(
      'list',
      help='List available modules of a pipeline.'
    )
    list_modules_p = add_common_args(list_modules_p)
    list_modules_p = add_pid_parser(list_modules_p)
    list_modules_p = add_json_parser(list_modules_p)

    # update module parser
    update_module_parser = action_parser.add_parser(
      'update',
      help='Update information associated with a module.'
    )
    update_module_parser = add_common_args(update_module_parser)
    update_module_parser = add_yaml_parser(update_module_parser)
    update_module_parser = add_force_parser(update_module_parser, 'module')

    return action_parser



"""Driver for AWS S3 compute"""

# Libs import
from basepair.modules.aws import S3

# App import
from .abstract import StorageAbstract

FILE_NOT_FOUND = 'file_not_found'
RESTORE_COMPLETE = 'restore_complete'
RESTORE_ERROR = 'restore_error'
RESTORE_IN_PROGRESS = 'restore_in_progress'
RESTORE_NOT_REQUIRED = 'restore_not_required'
RESTORE_NOT_STARTED = 'restore_not_started'


class Driver(StorageAbstract):
    """AWS S3 Driver class"""

    def __init__(self, cfg=None):
        """Instance constructor"""
        self.storage_settings = cfg.get('settings', {})
        self.s3_service = S3({
            'bucket': self.storage_settings.get('bucket'),
            'credentials': cfg.get('credentials'),
            'log_file': cfg.get('log_file'),
            'region': self.storage_settings.get('region'),
            'restore_period': self.storage_settings.get('restore_period'),
            'endpoint_url': self.storage_settings.get('endpoint_url'),
            'disable_sts': self.storage_settings.get('disable_sts', False),
        })

    def bulk_delete(self, uris):
        """Delete list of files by their uris"""
        return self.s3_service.bulk_delete(uris)

    def delete(self, uri):
        """Delete file from storage"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.delete(key)

    def download(self, uri, callback=None, file=None):
        """Download file from storage"""
        bucket = S3.get_bucket_from_uri(uri)
        key = S3.get_key_from_uri(uri)
        if callback:
            return self.s3_service.client.download_fileobj(
                bucket,
                key,
                file,
                Callback=callback,
            )
        return self.s3_service.download(key, file)

    def get_body(self, uri):
        """Get file body"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.get_file_body(key)

    def get_head(self, uri):
        """Get file head"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.get_object_head(key)

    def get_lifecycle(self, bucket=None):
        """Get storage lifecycle"""
        return self.s3_service.get_lifecycle(bucket)

    def get_public_url(self, uri):
        """Get a public accessible url"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.get_self_signed(key)

    def get_service(self):
        return self.s3_service

    def get_storage_context(self):
        """Get the storage context"""
        return self.s3_service.get_storage_context()

    def get_status(self, uri):
        """Get the file status"""
        bucket = S3.get_bucket_from_uri(uri)
        key = S3.get_key_from_uri(uri)
        return self.s3_service.get_status(bucket=bucket, key=key)

    def get_uri(self, key):
        """Get uri using key and storage settings"""
        return f's3://{self.s3_service.bucket}/{key}'

    def list(self, prefix, bucket=None):
        return self.s3_service.list(prefix, bucket)

    def restore_files_from_cold(self, uris, days):
        """Restore files from cold storage"""
        for uri in uris:
            status = self.get_status(uri)
            if status == 'restore_not_started':
                self.restore_from_cold(uri, days=days)

    def restore_from_cold(self, uri, days):
        """Restore file from cold storage"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.start_restore(key, days)

    def set_body(self, body, uri):
        """Set file body"""
        key = S3.get_key_from_uri(uri)
        return self.s3_service.set_file_body(body, key)

    def set_lifecycle(self, **kwargs):
        """Set storage lifecycle"""
        return self.s3_service.set_lifecycle(**kwargs)

    def upload(self, file_name, full_path, **kwargs):
        """Upload file to storage"""
        return self.s3_service.upload_file(file_name, full_path, **kwargs)

    def list_buckets(self):
        """List buckets"""
        return self.s3_service.list_buckets()



"""Driver for AWS Health Omics"""

# General imports
import time

# Libs import
from basepair.modules.aws import HOS
from basepair.modules.aws import S3

# App import
from .aws_s3 import Driver as S3Driver

RESTORE_ERROR = 'restore_error'
RESTORE_IN_PROGRESS = 'restore_in_progress'
RESTORE_NOT_REQUIRED = 'restore_not_required'
RESTORE_NOT_STARTED = 'restore_not_started'


class Driver(S3Driver):
    """AWS Health Omics Driver class"""

    def __init__(self, cfg=None):
        """Instance constructor"""
        super().__init__(cfg=cfg)  # calling parent class(S3) constructor to create the S3 client

        storage_settings = cfg.get('settings', {})
        self.hos_service = HOS({
            'credentials': cfg.get('credentials'),
            'reference_store_id': storage_settings.get('ho_reference_store_id'),
            'sequence_store_id': storage_settings.get('ho_sequence_store_id'),
            'region': storage_settings.get('ho_region'),
            'role_arn': storage_settings.get('ho_import_export_role_arn'),
        })

    def get_public_url(self, uri):
        """Get the public URL of the file"""
        if not self._is_health_omics_uri(uri):
            # If the URI is not a Health Omics URI, then call the S3 driver method
            return super().get_public_url(uri)

        # If the URI is a Health Omics URI, then get the public URL of the file with bucket.
        # Note: The bucket is an S3 access point URL not actual bucket name.
        # But presign will work with S3 access point URL as bucket name.
        bucket = S3.get_bucket_from_uri(uri)
        key = S3.get_key_from_uri(uri, bucket=bucket)
        return self.s3_service.get_self_signed(key, bucket=bucket)

    def get_service(self):
        return self.hos_service

    def get_storage_context(self):
        """Get the context for AWS HOS"""
        return {
            **super().get_storage_context(),
            'storage_driver': 'aws_hos'
        }

    def get_status(self, uri):
        """Get the file status"""
        if not self._is_health_omics_uri(uri):
            # If the URI is not a Health Omics URI, then call the S3 driver method
            return super().get_status(uri)

        # If the URI is a Health Omics URI, then get the read set metadata and return the status
        read_set_id = uri.split('/')[-2]
        metadata = self.hos_service.get_read_set_metadata(read_set_id)
        storage_status = metadata.get('status')

        if storage_status == 'ARCHIVED':
            return RESTORE_NOT_STARTED, storage_status
        elif storage_status == 'ACTIVE':
            return RESTORE_NOT_REQUIRED, storage_status
        elif storage_status == 'ACTIVATING':
            return RESTORE_IN_PROGRESS, storage_status
        else:
            return RESTORE_ERROR, storage_status

    def restore_from_cold(self, uri, days):
        """Restore file from cold storage"""
        if not self._is_health_omics_uri(uri):
            # If the URI is not a Health Omics URI, then call the S3 driver method
            return super().restore_from_cold(uri, days)

        # If the URI is a Health Omics URI, then start the read set activation job
        read_set_id = uri.split('/')[-2]
        max_attempts = 100
        attempt = 0
        while attempt < max_attempts:
            attempt += 1
            try:
                return self.hos_service.start_read_set_activation_job([read_set_id])
            except Exception:
                time.sleep(60)
        raise Exception("Failed to start read set activation job.")


    @staticmethod
    def _is_health_omics_uri(uri):
        """Check if uri is a S3 access point URL. Because Health Omics URIs are generated via S3 access points."""
        return '-s3alias' in uri



'''Host webapp api wrapper'''

# App imports
from .abstract import Abstract

class Host(Abstract):
  '''Webapp Host class'''
  def __init__(self, cfg):
    super(Host, self).__init__(cfg)
    self.endpoint += 'hosts/'



'''AWS Service abstract wrappers'''

# General imports
import json
import os
import time
from datetime import datetime

# Libs imports
import boto3
import requests

# App imports
from basepair.modules.logger import Logger
from .sts import STS

class Service: # pylint: disable=too-few-public-methods
  '''Abstract wrapper for services'''
  def __init__(self, cfg, service_name=''):
    self.cfg = cfg
    self.log = Logger.get_instance({'log_file': cfg.get('log_file')})
    self.session = None
    self.sts_service = self.connect()

    service = f'AWS {service_name} Service'
    self.handler = {
      'error': f'{service} Error:', # Blocking errors
      'info': f'{service} Info:', # Information
      'warning': f'{service} Warning:', # Non blocking errors
    }

  def connect(self):
    '''connect if session not valid anymore'''
    retry = 0
    # get session
    while True:
      session_args = Service.get_session_args(self.cfg, retry > 0)
      self.session = boto3.session.Session(**session_args)
      if self.cfg.get('disable_sts', False):
        return None
      else:
        sts_service = STS(self.session)
        if sts_service.is_valid_credential() or retry > 3:
          return sts_service
        retry += 1
        time.sleep(5) # else we sleep and try again

  def get_credentials(self):
    '''To get the current used credential for the session'''
    if hasattr(self, 'session'):
      client = self.session.get_credentials()
      return client.get_frozen_credentials()
    return None

  def get_log_msg(self, data):
    '''helper to return formatted log message'''
    default = {
      'exception': '',
      'msg': '',
      'log': True,
      'msg_type': 'error',
      'std_print': True,
    }
    data = {**default, **data}
    msg = f"{self.handler.get(data.get('msg_type'))}\n{data.get('msg')}"
    if data.get('std_print'):
      print(f"{msg}\n{data.get('exception') or ''}")
    if data.get('log') and hasattr(self.log, data.get('msg_type')):
      getattr(self.log, data.get('msg_type'))(msg, payload={'error': f"{data.get('exception')}"})
    return {
      'detail': data.get('exception'),
      'error': data.get('msg_type') == 'error',
      'msg': msg
    }

  @staticmethod
  def get_instance_meta_credentials(clean_cache=False):
    '''Getting credential from cache if exist'''
    tmp_file = '/tmp/.service_meta_info'
    credentials = {}
    expiration = None
    if not clean_cache:
      if os.path.isfile(tmp_file):
        with open(tmp_file, 'r', encoding='utf-8') as file:
          credentials = json.load(file)
          expiration = datetime.strptime(credentials.get('expiration'), '%Y-%m-%dT%H:%M:%SZ')

      if expiration and (expiration - datetime.utcnow()).total_seconds() > 60:
        return credentials

    try:
      os.unlink(tmp_file)
      instance_profile = requests.get('http://169.254.169.254/latest/meta-data/iam/security-credentials/')
      credentials = requests.get(f'http://169.254.169.254/latest/meta-data/iam/security-credentials/{instance_profile.text}').json()
      with open(tmp_file, 'w', encoding='utf-8') as file:
        file_content = json.dumps({
          'expiration': credentials.get('Expiration'),
          'id': credentials.get('AccessKeyId'),
          'secret': credentials.get('SecretAccessKey'),
          'token': credentials.get('Token'),
        })
        file.write(file_content)
    except Exception: # pylint: disable=broad-except
      credentials = {}
    return credentials

  @staticmethod
  def get_session_args(cfg, clean_cache=False):
    '''Generate the session arguments'''
    session_args = {}
    credential = cfg.get('credentials') or Service.get_instance_meta_credentials(clean_cache)
    if credential.get('assume_role'):
      sts_service = STS()
      sts_service.assume_role(credential['assume_role'])
      if sts_service.credential:
        credential['id'] = sts_service.credential.get('AccessKeyId')
        credential['secret'] = sts_service.credential.get('SecretAccessKey')
        credential['token'] = sts_service.credential.get('SessionToken')

    if credential.get('id'):
      session_args['aws_access_key_id'] = credential['id']

    if credential.get('secret'):
      session_args['aws_secret_access_key'] = credential['secret']

    if credential.get('token'):
      session_args['aws_session_token'] = credential['token']

    if cfg.get('region'):
      session_args['region_name'] = cfg['region']
    return session_args

  @staticmethod
  def get_tag_value(item, key):
    '''Helper to get tag value from item'''
    if isinstance(item, dict) and item.get('Tags'):
      return next(
        (tag for tag in item.get('Tags') if tag.get('Key') == key),
        {},
      ).get('Value') or None
    if hasattr(item, 'tags'):
      return next(
        (tag for tag in item.tags if tag.get('Key') == key),
        {},
      ).get('Value') or None
    return None



'''Analysis log wrapper'''

# General imports
import sys
import os

# Libs imports
from basepair.infra.webapp import Analysis

# App imports
from .abstract import Abstract


class AnalysisLog(Abstract):
  '''Analysis log class'''
  def __init__(self, analysis_id=None, config=None, kind=None, node_key=None):
    super().__init__(config, analysis_id)
    self.kind = kind
    self.node_key = node_key

  def get_data(self):
    '''Set log data'''
    return {
      'id': self.analysis_id,
      'log': {
        'bio': {
          self.node_key: {
            self.kind: {
              **self.log
            }
          }
        }
      }
    }

  def update_node(self, cmd=None, status=None, time_taken=None):
    '''Set node cmd, status'''
    node_data = {
      'status': status,
      'time_taken': time_taken
    }

    if cmd:
      node_data['commands'] = cmd
    data = {
      'id': self.analysis_id,
      'log': {
        'bio': {
          self.node_key: node_data
        }
      }
    }

    try:
      response = (Analysis(self.config.get('api'))).save_log(data)
      if response.get('error'):
        print('Error updating analysis node status')
        return
      print('Analysis node status updated')
    except Exception as error: #pylint: disable=broad-except
      print(f'ERROR: Updating analysis node status: {str(error)}')



"""Webapp API"""

# General imports
import json
import os

# Lib imports
import requests

# App imports
from basepair.helpers import eprint


class Abstract:
	"""Webapp abastract class"""

	def __init__(self, cfg):
		"""Constructor"""
		protocol = "https" if cfg.get("ssl", True) else "http"
		prefix = cfg.get("prefix")
		self.protocol = protocol
		self.host = cfg.get("host")
		self.endpoint = f"{protocol}://{self.host}{prefix}"
		self.payload = {"username": cfg.get("username"), "api_key": cfg.get("key")}
		self.headers = {"content-type": "application/json"}

	def delete(self, obj_id, verify=True):
		"""Delete resource"""
		try:
			response = requests.delete(
				f"{self.endpoint}{obj_id}",
				params=self.payload,
				timeout=5,
				verify=verify,
			)
			return self._parse_response(
				response,
				error_msgs=self._get_error_message(obj_id),
			)
		except requests.exceptions.RequestException as error:
			eprint(f"ERROR: {error}")
			return {"error": True, "msg": error}

	# pylint: disable=dangerous-default-value
	def get(
		self,
		obj_id,
		cache=False,
		params={},
		verify=True,
	):
		"""Get detail of an resource"""
		if not obj_id:
			return {"error": True, "msg": "Invalid or missing object id"}

		_cache = Abstract._get_from_cache(cache)
		if _cache:
			return _cache

		params.update(self.payload)
		try:
			response = requests.get(
				self.resource_url(obj_id),
				params=params,
				timeout=5,
				verify=verify,
			)
			parsed = self._parse_response(
				response,
				error_msgs=self._get_error_message(obj_id),
			)

			# save in cache if required
			Abstract._save_cache(cache, parsed)
			return parsed
		except requests.exceptions.RequestException as error:
			eprint(f"ERROR: {error}")
			return {"error": True, "msg": error}

	# pylint: disable=dangerous-default-value
	def list(
		self,
		cache=False,
		params={"limit": 100},
		verify=True,
	):
		"""Get a list of items"""
		_cache = Abstract._get_from_cache(cache)
		if _cache:
			return _cache

		params.update(self.payload)
		try:
			response = requests.get(
				self.endpoint.rstrip("/"),
				params=params,
				timeout=5,
				verify=verify,
			)
			parsed = self._parse_response(response)

			# save in cache if required
			Abstract._save_cache(cache, parsed)
			return parsed
		except requests.exceptions.RequestException as error:
			eprint(f"ERROR: {error}")
			return {"error": True, "msg": error}

	def list_all(self, filters={}):  # pylint: disable=dangerous-default-value
		"""Get a list of all items"""
		item_list = []
		limit = filters.get("limit", 100)  # default limit = 100
		offset = 0
		total_count = 1
		while len(item_list) < total_count:
			params = {"limit": limit, "offset": offset}
			params.update(filters)
			# response = self.list({**filters, 'limit': limit, 'offset': offset}) #TODO: Uncomment when everything moved to py3
			response = self.list(params=params)
			if response.get("error"):
				return {"error": True, "msg": response.get("msg")}
			total_count = response.get("meta", {}).get("total_count")
			item_list += response.get("objects")
			offset += limit
		return item_list

	def list_all_full(self, filters={}):
		"""Get all items but return full response format like list()"""
		item_list = []
		limit = filters.get("limit", 100)
		offset = 0
		total_count = 1
		while len(item_list) < total_count:
			params = {"limit": limit, "offset": offset}
			params.update(filters)
			# Reuse the list() method here
			response = self.list(params=params)
			if response.get("error"):
				return {"error": True, "msg": response.get("msg")}
			total_count = response.get("meta", {}).get("total_count", 0)
			item_list += response.get("objects", [])
			offset += limit

		meta = response.get("meta", {})
		meta["total_count"] = total_count
		meta["returned_count"] = len(item_list)
		return {"meta": meta, "objects": item_list}

	def resource_uri(self, obj_id):
		"""Generate resource uri from obj id"""
		return f"{self.pathname}{obj_id}"

	def resource_url(self, obj_id):
		"""Generate resource uri from obj id"""
		return f"{self.endpoint}{obj_id}"

	# pylint: disable=dangerous-default-value
	def save(
		self,
		obj_id=None,
		params={},
		payload={},
		verify=True,
	):
		"""Save or update resource"""
		params.update(self.payload)
		try:
			response = getattr(requests, "put" if obj_id else "post")(
				self.resource_url(obj_id) if obj_id else self.endpoint,
				data=json.dumps(payload),
				headers=self.headers,
				params=params,
				timeout=5,
				verify=verify,
			)
			return self._parse_response(
				response,
				error_msgs=self._get_error_message(obj_id),
			)
		except requests.exceptions.RequestException as error:
			eprint(f"ERROR: {error}")
			return {"error": True, "msg": error}

	@property
	def pathname(self):
		"""Get pathname"""
		return self.endpoint.replace(f"{self.protocol}://", "").replace(self.host, "")

	@staticmethod
	def _get_from_cache(cache):
		"""Helper to get data from cache"""
		if cache:
			filename = os.path.expanduser(cache)
			if os.path.exists(filename) and os.path.getsize(filename):
				with open(filename, "r", encoding="utf-8") as file:
					return json.loads(file.read().strip())
		return None

	@classmethod
	def _get_error_message(cls, obj_id):
		"""General response parser with obj id"""
		return {
			401: f"You don't have access to resource with id {obj_id}."
			if obj_id
			else "You don't have access to this resource.",
			404: f"Resource with id {obj_id} not found."
			if obj_id
			else "Resource not found.",
			500: "Error retrieving data from API!",
		}

	@classmethod
	def _parse_response(cls, response, error_msgs=None):
		"""General response parser"""
		error_msgs = error_msgs or {
			401: "You don't have access to this resource.",
			404: "Resource not found.",
			500: "Error retrieving data from API!",
		}
		if response.status_code in error_msgs:
			eprint(f"ERROR: {error_msgs[response.status_code]}")
			return {"error": True, "msg": error_msgs[response.status_code]}

		if response.status_code == 204:  # for delete response
			return {"error": False}

		try:
			response = response.json()
			error = isinstance(response, dict) and response.get("error")
			if error:
				if isinstance(error, dict):
					response = error
					msg = response.get("error_msgs")
					if msg:
						eprint(f"ERROR: {msg}")

					msg = response.get("warning_msgs")
					if msg:
						eprint(f"WARNING: {msg}")
				else:
					eprint(f"ERROR: {error}")

			return response
		except json.decoder.JSONDecodeError as error:
			msg = f"ERROR: Not able to parse response: {error}."
			eprint(msg)
			return {"error": True, "msg": msg}

	@staticmethod
	def _save_cache(cache, content):
		"""Helper to save the content in the cache"""
		if cache and content and not content.get("error"):
			filename = os.path.expanduser(cache)
			directory = os.path.dirname(filename)
			if not os.path.exists(directory):
				os.makedirs(directory)
			with open(filename, "w", encoding="utf-8") as handle:
				handle.write(json.dumps(content, indent=2))



'''AWS SQS Wrappers'''

# General import
import json

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class SQS(Service):
  '''Wrapper for SQS services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'SQS')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'sqs',
    })
    self.queue = cfg.get('queue')
    self.wait_time = cfg.get('wait_time', 20)

  def clear(self):
    '''Clear all message in queue'''
    self.client.purge_queue(QueueUrl=self.get_queue_url(self.queue))

  def create(self, queue):
    '''Create a queue'''
    try:
      self.client.create_queue(QueueName=queue)
      response = self.client.get_queue_url(QueueName=queue)
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to create queue {queue}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response.get('QueueUrl')

  def delete_message(self, handler):
    '''Delete a message from queue'''
    queue_url = self.get_queue_url(self.queue)
    if isinstance(queue_url, dict) and queue_url.get('error'):
      return queue_url

    try:
      response = self.client.delete_message(
        QueueUrl=queue_url,
        ReceiptHandle=handler
      )
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to remove message from queue {self.queue}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def get_message(self, visibility_time_out=1500, wait_time=None):
    '''Read one message from queue'''
    wait_time = wait_time or self.wait_time
    try:
      res = self.client.receive_message(
        QueueUrl=self.get_queue_url(self.queue),
        MaxNumberOfMessages=1,
        VisibilityTimeout=visibility_time_out,
        WaitTimeSeconds=wait_time,
      )
      message = res.get('Messages')[0]
      response = {
        'body': message.get('Body'),
        'handler': message.get('ReceiptHandle'),
        'id': message.get('MessageId'),
      }
    except (ClientError, IndexError) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get message from queue {self.queue}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def get_messages(self, limit=10, visibility_time_out=1500, wait_time=None):
    '''Read messages from queue'''
    wait_time = wait_time or self.wait_time
    try:
      res = self.client.receive_message(
        AttributeNames=['All'],
        QueueUrl=self.get_queue_url(self.queue),
        MaxNumberOfMessages=limit if limit <= 10 else 10,
        VisibilityTimeout=visibility_time_out,
        WaitTimeSeconds=wait_time,
      )
      response = [{
        'attributes': message.get('Attributes'),
        'body': message.get('Body'),
        'handler': message.get('ReceiptHandle'),
        'id': message.get('MessageId'),
      } for message in res.get('Messages')] if res.get('Messages') else []

    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get messages from queue {self.queue}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def get_queue_url(self, queue, create=True):
    '''Get the queue URL from queue name'''
    try:
      res = self.client.get_queue_url(QueueName=queue)
      response = res and res.get('QueueUrl')
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get queue url for queue {self.queue}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
      if create:
        self.get_log_msg({
          'exception': None,
          'msg': f'Creating new queue {self.queue}.',
          'msg_type': 'info',
        })
        self.create(queue)
        response = self.get_queue_url(queue, create=False)
    return response

  def send_message(self, msg, delay=0, queue=None):
    '''Send message'''
    queue = queue or self.queue
    queue_url = self.get_queue_url(queue)
    if isinstance(queue_url, dict) and queue_url.get('error'):
      return queue_url

    msg.pop('delay', None)
    if delay > 900:
      msg['delay'] = delay - 900
      delay = 900

    if isinstance(queue_url, str):
      try:
        response = self.client.send_message(
          DelaySeconds=delay,
          MessageBody=json.dumps(msg),
          QueueUrl=queue_url,
        )
      except ClientError as error:
        response = self.get_log_msg({
          'exception': error,
          'msg': f'Not able to send message to queue {queue}.',
        })
        if ExceptionHandler.is_throttled_error(exception=error):
          raise error
    return response

  def set_queue(self, queue):
    '''Set default queue'''
    self.queue = queue

  def update_message_visibility(self, handler, timeout):
    '''Update the message visibility timeout'''
    queue_url = self.get_queue_url(self.queue)
    try:
      self.client.change_message_visibility(
        QueueUrl=queue_url,
        ReceiptHandle=handler,
        VisibilityTimeout=timeout,
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': 'Not able to update the message visibility timeout.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error



'''Genome webapp api wrapper'''

# App imports
from .abstract import Abstract

class Genome(Abstract):
  '''Webapp Genome class'''
  def __init__(self, cfg):
    super(Genome, self).__init__(cfg)
    self.endpoint += 'genomes/'



import os
import json

from basepair.modules.secrets.drivers import SecretsAbstract


class Driver(SecretsAbstract):
  '''Local Driver for managing Secrets'''

  def __init__(self):
    '''Constructor for local secret driver class'''
    pass

  def get(self, secret_id, use_cache=True):
    '''Get the secrets using secret_id'''
    path = os.environ.get('SECRETS_PATH', '/home/ec2-user')
    secret_file_path = os.path.join(path, secret_id)
    if os.path.isfile(secret_file_path):
      secrets = ""
      with open(secret_file_path, 'r', encoding='utf-8') as file:
        secrets = json.load(file)
        return secrets.get('SecretString', '')



"""AWS S3 Wrappers"""

# General imports
import json
import mimetypes
import re
from datetime import datetime

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError, NoCredentialsError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

# Constants
DEFAULT_RESTORE_PERIOD = 7  # days
DEFAULT_TRANSITION_THRESHOLD = 30  # days


class S3(Service):  #  pylint: disable=too-many-public-methods
	"""Wrapper for S3 services"""

	def __init__(self, cfg):
		self.cfg = cfg
		super().__init__(cfg, "S3")
		self.bucket = cfg.get("bucket")
		client_vars = {
			"config": Config(
				retries={"max_attempts": 10, "mode": "standard"},
				signature_version="s3v4",
			),
			"service_name": "s3",
		}
		if cfg.get("endpoint_url"):
			client_vars["endpoint_url"] = cfg.get("endpoint_url")
		self.client = self.session.client(**client_vars)
		self.resource = self.session.resource("s3")

	def bulk_delete(self, uris):
		"""Bulk objects deletion by uri auto detecting bucket"""
		# validate uris
		uri_list = {"not_valid": {}, "valid": {}}
		for uri in uris:
			bucket = self.bucket
			validation_label = "not_valid"
			if uri.startswith("s3"):
				# group by bucket
				bucket = S3.get_bucket_from_uri(uri)
				validation_label = "valid"
			uri_list[validation_label][bucket] = (
				uri_list[validation_label].get(bucket) or []
			)
			uri_list[validation_label][bucket].append(
				{"Key": S3.get_key_from_uri(uri, bucket)}
			)

		# if there are valid uris we try to delete them
		response = self.get_log_msg(
			{
				"msg": "Nothing to delete.",
				"msg_type": "info",
			}
		)
		if uri_list.get("valid"):
			for bucket, objects in uri_list["valid"].items():
				try:
					self.client.delete_objects(
						Bucket=bucket,
						Delete={"Objects": objects},
					)
				except ClientError as error:
					response = self.get_log_msg(
						{
							"exception": error,
							"msg": "Not able to delete objects.",
						}
					)
					if ExceptionHandler.is_throttled_error(exception=error):
						raise error
				except Exception as error:
					response = self.get_log_msg(
						{
							"exception": error,
							"msg": "Not able to delete some objects.",
						}
					)
				else:
					response = self.get_log_msg(
						{
							"msg": "S3 deletion completed.",
							"msg_type": "info",
						}
					)

		# if there are invalid uris we return a warning
		if uri_list.get("not_valid"):
			response = self.get_log_msg(
				{
					"exception": "Not valid URI.",
					"msg": f"Not valid uris provided for deletion.\n{json.dumps(uri_list.get('not_valid'), indent=2)}",
					"msg_type": "warning",
				}
			)
		return response

	def delete(self, key, bucket=None):
		"""Delete object"""
		bucket = bucket or self.bucket
		try:
			self.client.delete_object(Bucket=bucket, Key=key)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to delete object key {key}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response
		return True

	def download(self, key, file=None, bucket=None):
		"""Download key as file"""
		bucket = bucket or self.bucket
		file = file or key.split("/")[-1]
		try:
			self.client.download_file(Bucket=self.bucket, Key=key, Filename=file)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to download object key {key}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response
		return True

	def get_bucket(self, bucket=None):
		"""Return bucket object"""
		bucket = bucket or self.bucket
		try:
			self.client.head_bucket(Bucket=bucket)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to access bucket {bucket}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response
		return self.resource.Bucket(bucket)

	def get_file_body(self, key):
		"""Return the content of the file if possible"""
		if self.get_object_head(key):
			try:
				response = self.client.get_object(
					Bucket=self.bucket,
					Key=key,
				)
				return response["Body"].read()
			except ClientError as error:
				self.get_log_msg(
					{
						"exception": error,
						"msg": f"Not able to get object body for key {key}.",
					}
				)
				if ExceptionHandler.is_throttled_error(exception=error):
					raise error
		return None

	def get_lifecycle(self, bucket=None):
		"""Get the current configuration of the object"""
		bucket = bucket or self.bucket
		response = None
		try:
			response = self.client.get_bucket_lifecycle_configuration(
				Bucket=bucket,
			)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to get lifecycle configuration for bucket {bucket}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return response

	def get_object_head(self, key, bucket=None, show_log=True):
		"""Check if key exists in S3 bucket"""
		bucket = bucket or self.bucket
		try:
			response = self.client.head_object(Bucket=bucket, Key=key)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to get object header for key {key} in bucket {bucket}.",
					"std_print": show_log,
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			response = int(error.response["Error"]["Code"]) != 404
		return response

	def get_self_signed(self, key, bucket=None, expires_in=28800, method="get_object"):
		"""Generate self signed url for key"""
		try:
			bucket = bucket or self.bucket
			return self.client.generate_presigned_url(
				method,
				Params={"Bucket": bucket, "Key": key},
				ExpiresIn=expires_in,  # a week
			)
		except (  # pylint: disable=broad-except
			ClientError,
			NoCredentialsError,
			Exception,
		) as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to self sign object key {key}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return None

	def get_storage_context(self):
		"""Method to return context for S3"""
		return {
			"storage_archival_enabled": bool(self.cfg.get("restore_period", False)),
			"storage_bucket": self.cfg.get("bucket"),
			"storage_driver": "aws_s3",
			"storage_region": self.cfg.get("region"),
			"storage_sse_enabled": "True",
			"storage_url": f"https://s3.{self.cfg.get('region')}.amazonaws.com",
		}

	def get_status(self, key, bucket=None):
		"""Method to check the status of file restore process"""
		status = "restore_error"
		storage_class = None
		head_response = self.get_object_head(key, bucket)
		if type(head_response).__name__ == "dict":
			storage_class = head_response.get("StorageClass") or "STANDARD"
			restore_status = head_response.get("Restore", "")
			# Object in STANDARD, INTELLIGENT_TIERING, GLACIER_IR, ONEZONE_IA, STANDARD_IA => Instant retrieval possible
			status = (
				"restore_not_started"
				if storage_class in ["DEEP_ARCHIVE", "GLACIER"]
				else "restore_not_required"
			)
			# check the status of restoration if had been triggered
			if restore_status and "true" in restore_status:
				status = "restore_in_progress"
			elif restore_status and "false" in restore_status:
				status = "restore_complete"
		elif head_response is False:  # 404 not found
			status = "file_not_found"
		return status, storage_class

	def get_storage_class(self, key, bucket=None):
		"""Return the storage class of the key"""
		response = self.get_object_head(key, bucket)
		return response.get("StorageClass", "") or "STANDARD"

	def list(self, prefix, bucket=None):
		"""List objects with prefix"""
		bucket = bucket or self.bucket
		try:
			all_files = []
			continuation_token = None
			while True:
				params = {"Bucket": bucket, "Prefix": prefix}
				if continuation_token:
					params["ContinuationToken"] = continuation_token

				response = self.client.list_objects_v2(**params)
				all_files.extend(response.get("Contents", []))
				if not response.get("IsTruncated"):  # At the end of the list?
					break
				continuation_token = response.get("NextContinuationToken")
			return all_files
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to list objects in bucket {bucket}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return {}

	def replicate(self, source, new_file, storage_class="STANDARD_IA"):
		"""Replicate a file from S3 to S3"""
		try:
			return self.client.copy_object(
				CopySource={
					"Bucket": source.get("bucket", self.bucket),
					"Key": source.get("key"),
				},
				Bucket=self.bucket,
				Key=new_file,
				StorageClass=storage_class,
			)
		except ClientError as error:
			self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to replicate object key {source}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return None

	def set_file_body(self, body, key):
		"""Write content to file"""
		try:
			return self.client.put_object(Body=body, Bucket=self.bucket, Key=key)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to set object body for key {key}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response

	def set_lifecycle(self, **kwargs):
		"""
		:param append_rule: (boolean) if previous rules are to be retained
		:param bucket: (string) bucket name for which the lifecycle rule is to be set
		:param prefix: (string) folder wise filter for which the lifecycle will be effective
		:param storage_class: (string) cold storage s3 class name where the files will be moved
		:param tag: (Array) of dictionary that will be used as a filter while applying lifecycle rule
								[{'Key': 'some_key', 'Value': 'some_value'}]
		"""
		transition_threshold = (
			kwargs.get("transition_threshold") or DEFAULT_TRANSITION_THRESHOLD
		)  # days
		storage_class = kwargs.get("storage_class") or "DEEP_ARCHIVE"
		prefix = kwargs.get("prefix") or "uploads/"
		tag = kwargs.get("tag")
		rules = []
		tag_name = tag and f'<{tag[0].get("Key")}:{tag[0].get("Value")}>'
		object_filter = (
			{"And": {"Prefix": prefix, "Tags": tag}} if tag else {"Prefix": prefix}
		)
		new_rule = {
			"Filter": object_filter,
			"ID": f'Move-files-in-{prefix}-{f"with-tag-{tag_name}" if tag else ""}-to-{storage_class}-\
				after-{transition_threshold}-days-<created-{datetime.now()}>',
			"Status": "Enabled",
			"Transitions": [
				{
					"Days": transition_threshold,
					"StorageClass": storage_class,
				},
			],
		}
		if kwargs.get("append_rule"):
			rules = self.convert_to_v2(self.get_lifecycle().get("Rules", []))
		rules.append(new_rule)
		bucket = kwargs.get("bucket") or self.bucket
		response = None
		try:
			response = self.client.put_bucket_lifecycle_configuration(
				Bucket=bucket,
				LifecycleConfiguration={
					"Rules": rules,
				},
			)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to update lifecycle configuration for bucket {bucket}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return response

	def start_restore(
		self, key, days=DEFAULT_RESTORE_PERIOD, bucket=None, tier="Standard"
	):
		"""
		Restore object using key
		If the object is not previously restored, then Amazon S3 returns 202 Accepted in the response.
		If the object is previously restored, Amazon S3 returns 200 OK in the response.
		DEEP_ARCHIVE only supports 'Standard' tier for retrieval
		"""
		restore_status, storage_class = self.get_status(key)
		if restore_status == "restore_not_started":
			try:
				self.client.restore_object(
					Bucket=bucket or self.bucket,
					Key=key,
					RestoreRequest={
						"Days": days,
						"GlacierJobParameters": {
							"Tier": "Expedited" if storage_class == "GLACIER" else tier
						},
					},
				)
				restore_status = "restore_in_progress"
			except ClientError as error:
				self.get_log_msg(
					{
						"exception": error,
						"msg": f"Not able to restore object key {key}.",
					}
				)
				if ExceptionHandler.is_throttled_error(exception=error):
					raise error
				restore_status = "restore_error"
		return restore_status, storage_class

		#  pylint: disable=too-many-arguments

	def upload_file(
		self,
		file_name,
		full_path,
		extra_args=None,
		force=False,
		show_log=True,
	):
		"""Upload a file to an S3 bucket"""
		extra_args = extra_args or {}
		mimetype, _ = mimetypes.guess_type(file_name)
		if mimetype:
			extra_args["ContentType"] = mimetype

		# Upload the file
		try:
			if force or not self.get_object_head(full_path, show_log=show_log):
				self.client.upload_file(
					file_name, self.bucket, full_path, ExtraArgs=extra_args
				)
			else:
				print(f"Skipping file {full_path} because already exist in S3.")
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to upload file {file_name}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response

		# get url by key
		return {
			"path": full_path,
			"url": self.get_self_signed(full_path),
		}

	def list_buckets(self):
		"""
		Lists all S3 buckets accessible
		"""
		try:
			response = self.client.list_buckets()
			buckets = response.get("Buckets", [])
			return [bucket["Name"] for bucket in buckets]

		except ClientError as error:
			print(f"Error listing buckets: {error}")
			raise error

	@staticmethod
	def convert_to_v2(rules):
		"""Method to convert the version1 lifecycle configuration format to v2"""
		for rule in rules:
			if "Prefix" in rule:
				rule["Filter"] = {"Prefix": rule["Prefix"]}
				rule.pop("Prefix")
		return rules

	@staticmethod
	def get_bucket_from_uri(uri):
		"""Helper to get bucket from uri"""
		match = re.search(r"^s3:\/\/([^:\/?\n]+)", uri)
		return match and match.group(1)

	@staticmethod
	def get_key_from_uri(uri, bucket=None):
		"""Helper to get key from uri"""
		bucket = bucket or S3.get_bucket_from_uri(uri)
		return uri.replace(f"s3://{bucket}/", "")



'''Genome datatype class'''

# General Import
import sys

# App imports
from bin.common_parser import add_common_args, add_uid_parser, add_json_parser

class Genome:
  '''Genome action methods'''

  @staticmethod
  def get_genome(bp_api, args):
    '''Get genome'''
    all_fail = True
    for uid in args.uid:
      all_fail = not bool(bp_api.print_data(data_type='genome', uid=uid, is_json=args.json)) and all_fail
    if all_fail:
      sys.exit('ERROR: Failed to load genome data.')

  @staticmethod
  def list_genome(bp_api, args):
    '''List genomes'''
    bp_api.print_data(data_type='genomes', is_json=args.json)

  @staticmethod
  def genome_action_parser(action_parser):
    '''Genome parser'''
    # get genome parser
    get_genome_p = action_parser.add_parser(
      'get',
      help='List detail info about one or more genomes.'
    )
    get_genome_p = add_common_args(get_genome_p)
    get_genome_p = add_uid_parser(get_genome_p, 'genome')
    get_genome_p = add_json_parser(get_genome_p)

    # list genome parser
    list_genomes_p = action_parser.add_parser(
      'list',
      help='List available genomes.'
    )
    list_genomes_p = add_common_args(list_genomes_p)
    list_genomes_p = add_json_parser(list_genomes_p)

    return action_parser



"""Alert abstract class"""


class AlertAbstract:
	"""Abstract class for log drivers"""

	def __init__(self):
		"""Constructor"""
		self.cfg = {}

	def critical(self, msg, payload):
		"""Send very high severity alert"""

	def high(self, msg, payload):
		"""Send high severity alert"""

	def info(self, msg, payload):
		"""Send very low severity alert"""

	def low(self, msg, payload):
		"""Send low severity alert"""

	def medium(self, msg, payload):
		"""Send medium severity alert"""

	def set_config(self, cfg):
		"""Helper to set configuration"""



"""Storage manager"""

# General imports
import importlib


class Storage:
    """Storage class"""
    def __init__(self, cfg):
        """Storage constructor"""
        driver = cfg.get('driver') or 'aws_s3'
        module = importlib.import_module(f'basepair.modules.storage.drivers.{driver}')
        self.driver = module.Driver(cfg)

    def bulk_delete(self, uris):
        """Delete list of files by their uris"""
        return self.driver.bulk_delete(uris)

    def delete(self, uri):
        """Delete file from storage"""
        return self.driver.delete(uri)

    def download(self, uri, callback=None, file=None):
        """Download file from storage"""
        return self.driver.download(uri, callback, file)

    def get_body(self, uri):
        """Get file body"""
        return self.driver.get_body(uri)

    def get_head(self, uri):
        """Get file head"""
        return self.driver.get_head(uri)

    def get_lifecycle(self, bucket=None):
        """Get storage lifecycle"""
        return self.driver.get_lifecycle(bucket)

    def get_public_url(self, uri):
        """Get a public accessible url"""
        return self.driver.get_public_url(uri)

    def get_service(self):
        """Get storage service object"""
        return self.driver.get_service()

    def get_status(self, uri):
        """Get the file status"""
        return self.driver.get_status(uri)

    def get_storage_context(self):
        """Get the storage context"""
        return self.driver.get_storage_context()

    def get_uri(self, key):
        """Get uri using key and storage settings"""
        return self.driver.get_uri(key)

    def list(self, prefix, bucket=None):
        """List files in prefix"""
        return self.driver.list(prefix, bucket)

    def restore_files_from_cold(self, uris, days):
        """Restore files from cold storage"""
        return self.driver.restore_files_from_cold(uris, days)

    def restore_from_cold(self, uri, days):
        """Restore file from cold storage"""
        return self.driver.restore_from_cold(uri, days)

    def set_body(self, body, uri):
        """Set file body"""
        return self.driver.set_body(body, uri)

    def set_lifecycle(self, **kwargs):
        """Set storage lifecycle"""
        return self.driver.set_lifecycle(kwargs)

    def upload(self, file_name, full_path, **kwargs):
        """Upload file to storage"""
        return self.driver.upload(file_name, full_path, **kwargs)



'''Analysis webapp api wrapper'''

# Lib imports
import json
import requests

# App imports
from basepair.helpers import eprint
from .abstract import Abstract

class Analysis(Abstract):
  '''Webapp Analysis class'''
  def __init__(self, cfg):
    super(Analysis, self).__init__(cfg)
    self.endpoint += 'analyses/'

  def bulk_start(self, payload={}, verify=True):
    '''Import sample from s3'''
    try:
      response = requests.post(
        '{}bulk_start'.format(self.endpoint),
        data=json.dumps(payload),
        headers=self.headers,
        params=self.payload,
        verify=verify,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}

  def reanalyze(self, payload={}, verify=True):
    '''Restart analysis'''
    try:
      response = requests.post(
        '{}reanalyze'.format(self.endpoint),
        data=json.dumps(payload),
        headers=self.headers,
        params=self.payload,
        verify=verify,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}

  def terminate(self, payload={}, verify=True):
    '''Terminate analysis'''
    try:
      response = requests.post(
        '{}terminate'.format(self.endpoint),
        data=json.dumps(payload),
        headers=self.headers,
        params=self.payload,
        verify=verify,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}

  def save_log(self, data):
    '''Save analysis log in db'''
    try:
      response = requests.post(
        '{}log'.format(self.endpoint),
        data=json.dumps(data),
        headers=self.headers,
        params=self.payload,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      print(f'ERROR: {error}')
      return {'error': True, 'msg': error}



#!/usr/bin/env python

from __future__ import print_function
from builtins import object
import sys
import json
import argparse
import time

import boto.swf
from boto.swf.exceptions import SWFDomainAlreadyExistsError, \
    SWFTypeAlreadyExistsError


class SWF(object):
    def __init__(self, config_file):
        self.conf = json.load(open(config_file))
        region = None
        try:
            for reg in boto.swf.regions():
                if reg.name == self.conf['aws']['swf']['region']:
                    region = reg
                    break

        except KeyError:
            region = None

        self.layer = boto.swf.layer1.Layer1(
            aws_access_key_id=self.conf['aws']['aws_id'],
            aws_secret_access_key=self.conf['aws']['aws_secret'],
            region=region
            )

        # self.s3conn = boto.connect_s3(
        #         aws_access_key_id=self.conf['aws']['aws_id'],
        #         aws_secret_access_key=self.conf['aws']['aws_secret'],
        #         )

    def terminate(self):
        workflows = self.layer.list_open_workflow_executions(
            self.conf['aws']['swf']['domain'], oldest_date=1
        )

        for workflow in workflows['executionInfos']:
            workflow_id = workflow['execution']['workflowId']
            self.layer.terminate_workflow_execution(
                self.conf['aws']['swf']['domain'], workflow_id)
            print(workflow_id, file=sys.stderr)

    def register(self):
        self.register_domain()
        self.register_workflow_type()
        self.register_activity_type()

    def register_domain(self):
        domain = self.conf['aws']['swf']['domain']

        try:
            self.layer.register_domain(domain, '7')
        except SWFDomainAlreadyExistsError:
            print('domain', domain, 'exists', file=sys.stderr)

        # print layer.list_domains('REGISTERED')

    def register_workflow_type(self):
        domain = self.conf['aws']['swf']['domain']
        workflow = self.conf['aws']['swf']['workflow_type']
        version = '1.0'

        try:
            self.layer.register_workflow_type(
                domain, workflow, version,
                task_list=self.conf['aws']['swf']['decider_task_list'],
                default_child_policy='REQUEST_CANCEL',
                default_execution_start_to_close_timeout='31536000',
                default_task_start_to_close_timeout='31536000',
            )
        except SWFTypeAlreadyExistsError:
            print('workflow', workflow, 'exists', file=sys.stderr)

    def register_activity_type(self):
        domain = self.conf['aws']['swf']['domain']
        activity = self.conf['aws']['swf']['activity_type']
        version = '1.0'

        try:
            self.layer.register_activity_type(
                domain, activity, version,
                default_task_heartbeat_timeout='31536000',
                default_task_schedule_to_close_timeout='31536000',
                default_task_schedule_to_start_timeout='31536000',
                default_task_start_to_close_timeout='31536000',
                task_list=self.conf['aws']['swf']['worker_task_list'],
            )
        except SWFTypeAlreadyExistsError:
            print('activity', activity, 'exists', file=sys.stderr)


def main():
    args = read_args()
    swf = SWF(args.config_file)

    if args.register:
        swf.register()

    if args.terminate:
        print('terminating ...', file=sys.stderr)
        swf.terminate()
        time.sleep(3)
        swf.terminate()


def read_args():
    parser = argparse.ArgumentParser(description='cmd line instance control')
    parser.add_argument('-t', '--terminate', action='store_true',
                        help='terminate open workflows')
    parser.add_argument('-r', '--register', action='store_true')
    parser.add_argument('-c', '--config-file')
    parser.set_defaults(
        config_file='setup/config.dev.json'
    )

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()



"""Driver for AWS S3 compute"""

# App import
from .aws_s3 import Driver as S3Driver
from .abstract import raise_no_implemented


class Driver(S3Driver):
    """Min IO Driver class"""
    def __init__(self, cfg=None):
        super().__init__(cfg)

    def get_lifecycle(self, bucket=None):
        """Get storage lifecycle"""
        raise_no_implemented("MinIO does not support lifecycle")

    def get_storage_context(self):
        """Get the context for Minio"""
        return {
            **super().get_storage_context(),
            'storage_driver': 'minio',
            'storage_url': self.storage_settings.get('endpoint_url'),
            'storage_sse_enabled': 'False'
        }

    def restore_from_cold(self, uri, days):
        """Restore file from cold storage"""
        raise_no_implemented("MinIO does not support restore from cold")

    def set_lifecycle(self, **kwargs):
        """Set storage lifecycle"""
        raise_no_implemented("MinIO does not support lifecycle")



"""AWS HealthOmics Storage Wrappers"""

# General imports
import uuid

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.service import Service


class HOS(Service):
    """Wrapper for Health Omics Storage services"""

    def __init__(self, cfg):
        super().__init__(cfg, 'HOS')
        self.client = self.session.client(**{
            'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
            'service_name': 'omics',
        })
        self.sequence_store_id = cfg.get('sequence_store_id')
        self.reference_store_id = cfg.get('reference_store_id')
        self.role_arn = cfg.get('role_arn')

    def batch_delete_read_set(self, read_set_ids):
        """Delete read sets"""
        try:
            response = self.client.batch_delete_read_set(sequenceStoreId=self.sequence_store_id, ids=read_set_ids)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to delete HealthOmics read sets: {str(error)}.',
            })
            raise error
        return response

    def delete_reference(self, reference_id):
        """Delete reference"""
        try:
            response = self.client.delete_reference(id=reference_id, referenceStoreId=self.reference_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to delete HealthOmics reference: {str(error)}.',
            })
            raise error
        return response

    def get_read_set_export_job(self, job_id):
        """Get export job"""
        try:
            response = self.client.get_read_set_export_job(id=job_id, sequenceStoreId=self.sequence_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to get HealthOmics export job information: {str(error)}.',
            })
            raise error
        return response

    def get_read_set_import_job(self, job_id):
        """Get import job"""
        try:
            response = self.client.get_read_set_import_job(id=job_id, sequenceStoreId=self.sequence_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to get HealthOmics import job information: {str(error)}.',
            })
            raise error
        return response

    def get_read_set_metadata(self, read_set_id):
        """Get read set metadata"""
        try:
            response = self.client.get_read_set_metadata(id=read_set_id, sequenceStoreId=self.sequence_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to get HealthOmics read set metadata: {str(error)}.',
            })
            raise error
        return response

    def get_reference_import_job(self, job_id):
        """Get reference import job"""
        try:
            response = self.client.get_reference_import_job(id=job_id, referenceStoreId=self.reference_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to get HealthOmics reference import job information: {str(error)}.',
            })
            raise error
        return response

    def get_reference_metadata(self, reference_id):
        """Get reference metadata"""
        try:
            response = self.client.get_reference_metadata(id=reference_id, referenceStoreId=self.reference_store_id)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to get HealthOmics reference metadata: {str(error)}.',
            })
            raise error
        return response

    def list_read_sets(self, filters):
        """Get read sets list"""
        try:
            response = self.client.list_read_sets(sequenceStoreId=self.sequence_store_id, filter=filters)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to list HealthOmics read sets: {str(error)}.',
            })
            raise error
        return response

    def list_references(self, filters):
        """Get references list"""
        try:
            response = self.client.list_references(referenceStoreId=self.reference_store_id, filter=filters)
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to list HealthOmics references: {str(error)}.',
            })
            raise error
        return response

    def start_read_set_activation_job(self, read_set_ids):
        """Start read set activation job"""
        try:
            sources = [{'readSetId': read_set_id} for read_set_id in read_set_ids]
            client_token = str(uuid.uuid4())
            return self.client.start_read_set_activation_job(
                sequenceStoreId=self.sequence_store_id,
                clientToken=client_token,
                sources=sources
            )
        except ClientError as error:
            self.get_log_msg({
                'exception': str(error),
                'msg': f'Not able to start HealthOmics read set activation job: {str(error)}.',
            })
            raise error

    def start_read_set_export_job(self, destination, sources):
        """Start export job"""
        try:
            client_token = str(uuid.uuid4())
            response = self.client.start_read_set_export_job(
                sequenceStoreId=self.sequence_store_id,
                destination=destination,
                roleArn=self.role_arn,
                clientToken=client_token,
                sources=sources
            )
        except ClientError as error:
            self.get_log_msg({
                'exception': error,
                'msg': f'Not able to start HealthOmics export job: {str(error)}.',
            })
            raise error
        return response

    def start_read_set_import_job(self, sources):
        """Start import job"""
        try:
            client_token = str(uuid.uuid4())
            return self.client.start_read_set_import_job(
                sequenceStoreId=self.sequence_store_id,
                roleArn=self.role_arn,
                clientToken=client_token,
                sources=sources
            )
        except ClientError as error:
            self.get_log_msg({
                'exception': str(error),
                'msg': f'Not able to start HealthOmics import job: {str(error)}.',
            })
            raise error

    def start_reference_import_job(self, sources):
        """Start reference import job"""
        try:
            client_token = str(uuid.uuid4())
            return self.client.start_reference_import_job(
                referenceStoreId=self.reference_store_id,
                roleArn=self.role_arn,
                clientToken=client_token,
                sources=sources
            )
        except ClientError as error:
            self.get_log_msg({
                'exception': str(error),
                'msg': f'Not able to start HealthOmics reference import job: {str(error)}.',
            })
            raise error



'''AWS STS wrappers'''

# General imports
from datetime import datetime

# Libs imports
import boto3
from botocore.client import Config
from botocore.exceptions import ClientError

# App imports
from basepair.modules.logger import Logger

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler

class STS: # pylint: disable=too-few-public-methods
  '''Abstract wrapper for services'''
  def __init__(self, session=None, service_name='GeneralPurpose'):
    self.client = session.client(**{
      'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'sts',
    }) if session else boto3.client('sts')
    self.credential = None
    self.log = Logger.get_instance()
    self.service_name = service_name

    service = f'AWS {service_name} Service'
    self.handler = {
      'error': f'{service} Error:', # Blocking errors
      'info': f'{service} Info:', # Information
      'warning': f'{service} Warning:', # Non blocking errors
    }

  def assume_role(self, role):
    '''Assume role'''
    now = datetime.now()
    args = {
      "RoleArn": role,
      "RoleSessionName":f'AssummingRoleFor{self.service_name}_{now.strftime("%Y%m%d%H%M%S")}',
    }
    try:
      # Note: This will only work if in the role definition
      # the session timeout is larger than the requested
      caller_identity = self.client.get_caller_identity()
      if "assumed-role" not in caller_identity.get('Arn'):
        args['DurationSeconds'] = 43200
      response = self.client.assume_role(**args)
      self.credential = response.get('Credentials')
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to assume role {role}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

  def is_valid_credential(self):
    '''Check if valid credential'''
    try:
      self.client.get_caller_identity()
      return True
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': 'Not able to get caller identity.',
        'msg_type': 'warning',
      })
      return False

  def get_log_msg(self, data):
    '''helper to return formated log message'''
    default = {
      'exception': '',
      'msg': '',
      'log': True,
      'msg_type': 'error',
      'std_print': True,
    }
    data = {**default, **data}
    msg = f"{self.handler.get(data.get('msg_type'))}\n{data.get('msg')}"
    if data.get('std_print'):
      print(f"{msg}\n{data.get('exception') or ''}")
    if data.get('log') and hasattr(self.log, data.get('msg_type')):
      getattr(self.log, data.get('msg_type'))(msg, payload={'error': f"{data.get('exception')}"})
    return {
      'detail': data.get('exception'),
      'error': data.get('msg_type') == 'error',
      'msg': msg
    }



'''AWS CW Wrappers'''

# General import
from datetime import datetime, timedelta
from operator import itemgetter

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class CW(Service): # pylint: disable=too-few-public-methods
  '''Wrapper for CW services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'CW')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'cloudwatch',
    })

  def get_cpu_load_avg(self, instance_id, range_in_seconds=300):
    '''Get cpu load average'''
    now = datetime.utcnow()
    past = now - timedelta(seconds=range_in_seconds)
    try:
      res = self.client.get_metric_statistics(
        Namespace='AWS/EC2',
        MetricName='CPUUtilization',
        Dimensions=[{'Name': 'InstanceId', 'Value': instance_id}],
        StartTime=past,
        EndTime=now,
        Period=60,
        Statistics=['Average']
      )
      last_datapoint = sorted(res.get('Datapoints'), key=itemgetter('Timestamp'))
      return len(last_datapoint) > 0 and last_datapoint[-1].get('Average')
    except IndexError as error:
      detail = error
      msg = 'No metrics available.'
    except ClientError as error:
      detail = error
      msg = 'Failing trying to access metrics.'
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return self.get_log_msg({'exception': detail, 'msg': msg})

  def get_mem_load_avg(self, instance, range_in_seconds=300):
    '''Get mem load average'''
    now = datetime.utcnow()
    past = now - timedelta(seconds=range_in_seconds)
    try:
      res = self.client.get_metric_statistics(
        Namespace='CWAgent',
        MetricName='mem_used_percent',
        Dimensions=[{
          'Name': 'InstanceId',
          'Value': instance.id,
        }, {
          'Name': 'ImageId',
          'Value': instance.image_id,
        }, {
          'Name': 'InstanceType',
          'Value': instance.instance_type,
        }],
        StartTime=past,
        EndTime=now,
        Period=60,
        Statistics=['Average']
      )
      last_datapoint = sorted(res.get('Datapoints'), key=itemgetter('Timestamp'))
      return len(last_datapoint) > 0 and last_datapoint[-1].get('Average')
    except IndexError as error:
      detail = error
      msg = 'No metrics available.'
    except ClientError as error:
      detail = error
      msg = 'Failing trying to access metrics.'
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return self.get_log_msg({'exception': detail, 'msg': msg})

  def get_storage_load_avg(self, instance, range_in_seconds=300):
    '''Get disk used average'''
    now = datetime.utcnow()
    past = now - timedelta(seconds=range_in_seconds)
    try:
      res = self.client.get_metric_statistics(
        Namespace='CWAgent',
        MetricName='disk_used_percent',
        Dimensions=[{
          'Name': 'InstanceId',
          'Value': instance.id,
        }, {
          'Name': 'ImageId',
          'Value': instance.image_id,
        }, {
          'Name': 'InstanceType',
          'Value': instance.instance_type,
        }, {
          'Name': 'device',
          'Value': 'nvme0n1p1',
        }, {
          'Name': 'fstype',
          'Value': 'ext4',
        }, {
          'Name': 'path',
          'Value': '/',
        }],
        StartTime=past,
        EndTime=now,
        Period=60,
        Statistics=['Average']
      )
      last_datapoint = sorted(res.get('Datapoints'), key=itemgetter('Timestamp'))
      return len(last_datapoint) > 0 and last_datapoint[-1].get('Average')
    except IndexError as error:
      detail = error
      msg = 'No metrics available.'
    except ClientError as error:
      detail = error
      msg = 'Failing trying to access metrics.'
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return self.get_log_msg({'exception': detail, 'msg': msg})


"""set up basepair package"""
from __future__ import print_function
import os
import sys

import requests

from .utils import colors

# Exposing infra webapp library
from .infra.webapp import (
	Analysis,
	File,
	Gene,
	Genome,
	GenomeFile,
	Host,
	Module,
	Pipeline,
	Project,
	Sample,
	Upload,
	User,
)

# Exposing the storage wrapper

__title__ = "basepair"
__version__ = "2.4.0"
__copyright__ = "Copyright [2017] - [2025] Basepair INC"


JSON_URL = f"https://pypi.python.org/pypi/{__title__}/json"

if not os.environ.get("SECRETS_DRIVER") == "local":
	try:
		resp = requests.get(JSON_URL, timeout=1)
		if resp.status_code == 200:
			latest_version = resp.json()["info"]["version"]
			if latest_version != __version__:
				print(
					colors.color.warning(
						f"WARNING: The latest version of basepair package is {latest_version}. "
						"Please upgrade to avail the latest features and bug-fixes"
					),
					file=sys.stderr,
				)
	except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
		print("warning: no internet", file=sys.stderr)


def connect(*args, **kwargs):
	"""return basepair package"""
    # pylint: disable=import-outside-toplevel
	from basepair.api import BpApi
	return BpApi(*args, **kwargs)



"""Alert factory"""

# General imports
import importlib

INSTANCES = {}


class Alert:  # pylint: disable=too-few-public-methods
	"""Alert factory class"""

	@staticmethod
	def get_instance(cfg=None):  # pylint: disable=dangerous-default-value
		"""Alert instantiation"""
		if cfg is None:
			cfg = {}
		driver = cfg.get("driver")
		driver_module = importlib.import_module(
			f"basepair.modules.alert.drivers.{driver}"
		)
		INSTANCES[driver] = INSTANCES.get(driver) or driver_module.Instance()
		INSTANCES[driver].set_config(cfg)
		return INSTANCES[driver]



'''AWS Exception handler class'''

class ExceptionHandler:
  '''Helper class'''
  def __init__(self):
    pass

  @staticmethod
  def is_insufficient_capacity_error(exception=None):
    '''Check if a insufficient capacity error'''
    if exception:
      return (getattr(exception, 'response', {})).get('Error', {}).get('Code') == 'InsufficientInstanceCapacity'
    return None

  @staticmethod
  def is_throttled_error(exception=None):
    '''Check if a throttled error'''
    if exception:
      return (getattr(exception, 'response', {})).get('Error', {}).get('Code') in ExceptionHandler.get_throttled_error_codes()
    return None

  @staticmethod
  def get_throttled_error_codes():
    '''Get a list of all throttled error codes'''
    return [
      'Throttling',
      'ThrottlingException',
      'ThrottledException',
      'RequestThrottledException',
      'TooManyRequestsException',
      'ProvisionedThroughputExceededException',
      'TransactionInProgressException',
      'RequestLimitExceeded',
      'BandwidthLimitExceeded',
      'LimitExceededException',
      'ReadTimeoutError',
      'RequestThrottled',
      'SlowDown',
      'PriorRequestNotComplete',
      'EC2ThrottledException',
    ]



"""Driver for Slack"""

# General imports
import json

# Libs import
from slack_sdk import WebClient


# App imports
from .abstract import AlertAbstract


class Instance(AlertAbstract):
	"""Slack Implementation of Abstract class"""

	channel = "#alerts"
	client = None
	priority_emojis = {
		"P1": ":rotating_light:",  # Critical
		"P2": ":bangbang:",  # High
		"P3": ":warning:",  # Medium
		"P4": ":information_source:",  # Low
		"P5": ":speech_balloon:",  # Info
	}
	username = "AlertBot"
	token = ""

	def critical(self, msg, payload, channel=None):
		"""Send critical severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, channel, "P1"))

	def high(self, msg, payload, channel=None):
		"""Send high severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, channel, "P2"))

	def info(self, msg, payload, channel=None):
		"""Send very low severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, channel, "P5"))

	def low(self, msg, payload, channel=None):
		"""Send low severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, channel, "P4"))

	def medium(self, msg, payload, channel=None):
		"""Send medium severity alert"""
		return self._send_alert(self._parse_payload(msg, payload, channel, "P3"))

	def set_config(self, cfg):
		"""Set Slack webhook config"""
		self.channel = cfg.get("channel") or self.channel
		self.token = cfg.get("token")
		self.username = cfg.get("username") or self.username
		self.client = WebClient(token=self.token)

	def _parse_payload(self, msg, payload, channel=None, priority="P3"):
		"""Parse payload"""
		emoji = self.priority_emojis.get(priority, ":grey_question:")
		payload_str = json.dumps(payload or {}, indent=2)
		return {
			"channel": channel or self.channel,
			"text": f"{emoji} *{priority}* - {msg}\n```json\n{payload_str}\n```",
			"username": self.username,
		}

	def _send_alert(self, payload):
		"""Send alert"""
		return self.client.chat_postMessage(**payload)
